part of '../flutterx_live_data.dart';

class MutableLiveData<T> extends LiveData<T> {
  set value(T value) => _setData(value);

  MutableLiveData({T? value, super.equality}) : assert(value is T, 'value must be of type $T, got $value') {
    _setData(value as T);
  }

  MutableLiveData.late({super.equality});

  void postValue(FutureOr<T> value) => _postSetData(value);
}

class MediatorMutableLiveData<T> extends MutableLiveData<T> with MediatorMixin<T> {
  MediatorMutableLiveData({super.value, super.equality, bool reactive = false}) {
    this.reactive = reactive;
  }

  MediatorMutableLiveData.late({super.equality, bool reactive = false}) : super.late() {
    this.reactive = reactive;
  }

  MediatorMutableLiveData.copy(ObservableData<T> source, {bool reactive = false}) : super.late() {
    this.reactive = reactive;
    addSource<T>(source, _setData);
  }
}

class HistoryMutableLiveData<T> extends MutableLiveData<T> with HistoryMixin<T> {
  @override
  final StringEncode<T>? encode;

  @override
  final StringDecode<T>? decode;

  HistoryMutableLiveData({super.value, this.encode, this.decode});

  HistoryMutableLiveData.late({this.encode, this.decode}) : super.late();
}

class StorageMutableLiveData<T> extends MutableLiveData<T> with StorageMixin<T> {
  static final String _empty = String.fromCharCode(0);
  @override
  final String key;

  @override
  final T defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final StringEncode<T> encode;

  StorageMutableLiveData({
    required this.key,
    T? defaultValue,
    required this.storage,
    this.encode = jsonEncode,
    required StringDecode<T> decode,
  })  : assert(defaultValue is T, 'defaultValue must be of type $T, got $defaultValue'),
        defaultValue = defaultValue as T,
        super(value: StorageMixin.resolve(key, defaultValue, storage, decode));

  StorageMutableLiveData.string({
    required String key,
    T? defaultValue,
    required Map<String, String> storage,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: (value) => value?.toString() ?? _empty,
            decode: (source) => (source == _empty ? null : source) as T);

  StorageMutableLiveData.int({
    required String key,
    T? defaultValue,
    required Map<String, String> storage,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: (value) => value?.toString() ?? '',
            decode: (source) => (source.isEmpty ? null : int.parse(source)) as T);

  StorageMutableLiveData.double({
    required String key,
    T? defaultValue,
    required Map<String, String> storage,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: (value) => value?.toString() ?? '',
            decode: (source) => (source.isEmpty ? null : double.parse(source)) as T);

  StorageMutableLiveData.bool({
    required String key,
    T? defaultValue,
    required Map<String, String> storage,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: (value) => value?.toString() ?? '',
            decode: (source) => (source.isEmpty ? null : ('true' == source)) as T);

  StorageMutableLiveData.enumeration({
    required String key,
    T? defaultValue,
    required Map<String, String> storage,
    required Iterable<T> values,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: StorageMixin.encodeEnum,
            decode: (source) => StorageMixin.decodeEnum(source, values));

  StorageMutableLiveData.jsonObject({
    required String key,
    T? defaultValue,
    required Map<String, String> storage,
    dynamic Function(T value)? toJson,
    required T Function(Map<String, dynamic> json) fromJson,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: (value) => jsonEncode(toJson == null ? value : toJson(value)),
            decode: (source) => fromJson(jsonDecode(source)));

  StorageMutableLiveData.jsonList({
    required String key,
    T? defaultValue,
    required Map<String, String> storage,
    dynamic Function(T value)? toJson,
    required T Function(List json) fromJson,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: (value) => jsonEncode(toJson == null ? value : toJson(value)),
            decode: (source) => fromJson(jsonDecode(source)));
}
