part of '../flutterx_live_data.dart';

mixin HistoryMixin<T> on ObservableData<T> {
  int _capacity = 20;
  int _index = -1;
  bool _sync = false;
  @protected
  final Queue<String> queue = Queue();
  final LiveData<bool> canUndo = MutableLiveData(value: false);
  final LiveData<bool> canRedo = MutableLiveData(value: false);

  bool get _canUndo => _index > 0;

  bool get _canRedo => _index < queue.length - 1;

  int get capacity => _capacity;

  set capacity(int value) {
    if (value < 1) throw ArgumentError.value(value, 'capacity', 'Cannot be less than 1');
    _capacity = value;
    if (queue.length >= value) {
      queue.removeFirst();
      while (queue.length >= value) queue.removeFirst();
      if (_index >= queue.length)
        index = queue.length - 1;
      else
        updateValues();
    }
  }

  int get index => _index;

  set index(int value) {
    if (value == _index) return;
    if (value < -1 || value >= queue.length)
      throw IndexError.withLength(value, queue.length, indexable: queue, name: 'index');
    _index = value;
    if (!_sync && value >= 0) restoreState();
    updateValues();
  }

  int get length => queue.length;

  @protected
  StringEncode<T>? get encode => null;

  @protected
  StringDecode<T>? get decode => null;

  @override
  void invalidate() {
    if (!_sync) saveState();
    super.invalidate();
  }

  void historyUndo() {
    if (!_canUndo) throw StateError('Cannot undo');
    index--;
  }

  void historyRedo() {
    if (!_canRedo) throw StateError('Cannot redo');
    index++;
  }

  void historyDelete() {
    if (queue.isEmpty) throw StateError('Cannot delete');
    while (queue.length - 1 >= _index) queue.removeLast();
    index--;
  }

  void historyReplace(T data, {int? index}) {
    if (queue.isEmpty) throw StateError('Cannot replace');
    index ??= _index;
    if (index > _index) throw IndexError.withLength(index, queue.length, indexable: queue, name: 'index');
    while (queue.length - 1 >= index) queue.removeLast();
    _index = index - 1;
    _data = data;
    invalidate();
  }

  void historyClear() {
    queue.clear();
    index = -1;
  }

  @protected
  @mustCallSuper
  void saveState() {
    try {
      _sync = true;
      final state = (encode ?? jsonEncode)(_data);
      final index = _index;
      if (index >= 0 && state == queue.elementAt(index)) return;
      while (index < queue.length - 1) queue.removeLast();
      queue.add(state);
      if (index + 1 >= _capacity)
        queue.removeFirst();
      else {
        this.index++;
      }
    } finally {
      _sync = false;
    }
  }

  @protected
  @mustCallSuper
  void restoreState() {
    try {
      _sync = true;
      _data = (decode ?? jsonDecode)(queue.elementAt(_index));
      invalidate();
    } finally {
      _sync = false;
    }
  }

  @protected
  void updateValues() {
    canUndo._setData(_canUndo);
    canRedo._setData(_canRedo);
  }
}
