part of '../flutterx_live_data.dart';

mixin MediatorMixin<Y> on ObservableData<Y> {
  final Map<ObservableData, _Source> _sources = {};
  bool _reactive = false;

  @nonVirtual
  @protected
  bool get reactive => _reactive;

  @nonVirtual
  @protected
  set reactive(bool value) {
    if (value == _reactive) return;
    _reactive = value;
    if (!active) value ? _plugAll() : _unplugAll();
  }

  @override
  void onActive() {
    super.onActive();
    if (!_reactive) _plugAll();
  }

  @override
  void onInactive() {
    if (!_reactive) _unplugAll();
    super.onInactive();
  }

  void addSource<X>(ObservableData<X> source, Observer<X> onChanged) {
    assert(!_sources.containsKey(source), 'source already registered');
    final wrapper = _Source(() => source.observeForever(onChanged), () => source.removeObserver(onChanged));
    _sources[source] = wrapper;
    if (active || _reactive) wrapper.plug();
  }

  void removeSource(ObservableData source) {
    final wrapper = _sources.remove(source);
    assert(wrapper != null, 'source $source not registered');
    if (active || _reactive) wrapper?.unplug();
  }

  void removeSources(Iterable<ObservableData> sources) => [...sources].forEach(removeSource);

  void invalidateSource(ObservableData source) {
    if (active || _reactive) {
      final wrapper = _sources[source];
      assert(wrapper != null, 'source $source not registered');
      wrapper?.invalidate();
    }
  }

  void invalidateSources([Iterable<ObservableData>? sources]) {
    if (active || _reactive) {
      if (sources == null)
        for (final source in _sources.values) source.invalidate();
      else
        [...sources].forEach(invalidateSource);
    }
  }

  void initialize(ValueGetter<FutureOr<Y>> initializer,
          {ObservableData invalidate = const ImmutableLiveData<void>(null)}) =>
      addSource(invalidate, (_) => _postSetData(initializer()));

  void addTransformSources(Iterable<ObservableData> sources, FutureOr<Y> Function(List values) transform) =>
      ObservableData.observeAll(sources, (values) => _postSetData(transform(values)), addSource);

  void addTransformSource<X>(ObservableData<X> source, FutureOr<Y> Function(X) transform) =>
      addSource<X>(source, (value) => _postSetData(transform(value)));

  Disposable addLiveTransformSources(
      Iterable<ObservableData> sources, ObservableData<Y> Function(List values) transform) {
    ObservableData<Y>? lastSource;
    ObservableData.observeAll(
        sources, (values) => lastSource = _applyLiveTransform<List>(values, lastSource, transform), addSource);
    return _Disposable(() {
      removeSources(sources);
      if (lastSource != null) removeSource(lastSource!);
    });
  }

  Disposable addLiveTransformSource<X>(ObservableData<X> source, ObservableData<Y> Function(X) transform) {
    ObservableData<Y>? lastSource;
    addSource<X>(source, (value) => lastSource = _applyLiveTransform<X>(value, lastSource, transform));
    return _Disposable(() {
      removeSource(source);
      if (lastSource != null) removeSource(lastSource!);
    });
  }

  ObservableData<Y> _applyLiveTransform<T>(
      T value, ObservableData<Y>? lastSource, ObservableData<Y> Function(T value) transform) {
    final newSource = transform(value);
    if (!identical(lastSource, newSource)) {
      if (lastSource != null) removeSource(lastSource);
      addSource(newSource, _setData);
    }
    return newSource;
  }

  void _plugAll() {
    for (final source in [..._sources.values]) source.plug();
  }

  void _unplugAll() {
    for (final source in [..._sources.values]) source.unplug();
  }
}

@immutable
class _Source {
  final VoidCallback plug;
  final VoidCallback unplug;

  const _Source(this.plug, this.unplug);

  void invalidate() {
    unplug();
    plug();
  }
}
