part of '../flutterx_live_data.dart';

abstract class ObservableData<T> {
  final Map<Observer<T>, ObserverWrapper<T>> _observers = {};
  final Equality<T> equality;
  late T _data;
  int _version = -1;
  int _activeCount = 0;

  int get version => _version;

  bool get initialized => _version > -1;

  bool get active => _activeCount > 0;

  @protected
  ObservableData({Equality<T>? equality}) : equality = equality ?? const Equality();

  ObserverWrapper<T> addObserverWrapper(ObserverWrapper<T> observer) {
    assert(!_observers.containsKey(observer.onChanged), 'observer already registered');
    return _observers[observer.onChanged] ??= observer.._attach(this);
  }

  void removeObserverWrapper(ObserverWrapper<T> observer) => removeObserver(observer.onChanged);

  ObserverWrapper<T> observeForever(Observer<T> onChanged, {Dispatcher? dispatcher}) =>
      addObserverWrapper(ObserverWrapper<T>(onChanged, dispatcher));

  void removeObserver(Observer<T> onChanged) {
    assert(_observers.containsKey(onChanged), 'observer not registered');
    _observers.remove(onChanged)?._detach(this);
  }

  @protected
  void onActive() {}

  @protected
  void onInactive() {}

  @protected
  void _postSetData(FutureOr<T> value) => value is Future<T> ? value.then(_setData) : _setData(value);

  @protected
  void _setData(T value) {
    if (!initialized || !equality.equals(_data, value)) {
      _data = value;
      invalidate();
    }
  }

  @protected
  @mustCallSuper
  void invalidate() {
    _version++;
    for (final observer in [..._observers.values]) observer._considerNotify();
  }

  @override
  String toString() => describe(this);

  void _onObserverAttach(ObserverWrapper<T> observer) {
    if (++_activeCount == 1) onActive();
  }

  void _onObserverDetach(ObserverWrapper<T> observer) {
    if (--_activeCount == 0) onInactive();
  }

  static void observeAll(
    Iterable<ObservableData> sources,
    void Function(List values) onChanged,
    void Function<V>(ObservableData<V> source, Observer<V> onChanged) observe,
  ) {
    final data = [...sources];
    var ready = false;
    void observeAll(dynamic _) {
      if (ready) {
        List values;
        try {
          values = _allValues(data);
        } catch (_) {
          return; // at least one source is not initialized
        }
        onChanged(values);
      }
    }

    final length = data.length;
    final last = length - 1;
    for (var i = 0; i < length; i++) {
      if (i == last) ready = true;
      observe(data[i], observeAll);
    }
  }

  static String describe(ObservableData data) =>
      '${objectRuntimeType(data, 'ObservableData')}${data.initialized ? (identical(data._data, data) ? '' : '(${data._data})') : '[uninitialized]'}';

  static List _allValues(Iterable<ObservableData> sources) => [...sources.map((data) => data._data)];
}
